/* C */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <malloc.h>

/* Windows */
#include <wtypes.h>
#include <wincon.h>

/* Local */
#include "winmutt.h"
#include "wmuttrc.h"
#include "wmutil.h"
#include "wmvars.h"
#include "wmbox.h"
#include "wmprocess.h"

/* close down - tidy up */

void wm_Exit(int errLvl) {
  struct wmAlias *thisAlias, *nextAlias;

  /* free alias data */
  thisAlias = wmFirstAlias;

  while(thisAlias != NULL) {
    free(thisAlias->aliasAddr);
    free(thisAlias->aliasName);
    nextAlias = thisAlias->nextAlias;
    free(thisAlias);
    thisAlias = nextAlias;
  }

  /* ...
     free set strings too
     and mails
     ... */

  if(errLvl) wm_getPress();

  FreeConsole();

  exit(errLvl);
}


/* initialise various variables etc */
void initVars() {
  CONSOLE_SCREEN_BUFFER_INFO csbi;
  int i;

  wmFirstAlias = NULL;
  wmLastAlias = NULL;

  wmFirstMail = NULL;
  wmLastMail = NULL;

  GetConsoleScreenBufferInfo(wmConOut,&csbi);

  WMCOL_FGDEFAULT = csbi.wAttributes & (WMCOL_FGWHITE | WMCOL_FGBRIGHT); /* ie, extract the fg colours */
  WMCOL_BGDEFAULT = csbi.wAttributes & (WMCOL_BGWHITE | WMCOL_BGBRIGHT); /* likewise bg */

  /* initialise all the colours to the current console colours */
  for(i=0;i < WMNUMOFCOLS;i++) {
    wmColours[i].colourNum = WMCOL_FGDEFAULT | WMCOL_BGDEFAULT;
    wmColours[i].colPattern = NULL;
  }
  /* invert the indicator colour (preserving brightness) */
  wmColours[WMINDICATORCOL].colourNum = (WMCOL_FGWHITE ^ (WMCOL_FGDEFAULT & WMCOL_FGWHITE)) |
                                        (WMCOL_BGWHITE ^ (WMCOL_BGDEFAULT & WMCOL_BGWHITE)) |
                                        (csbi.wAttributes & (WMCOL_FGBRIGHT | WMCOL_BGBRIGHT));

  wmQuitting = FALSE;
  wmWasErr = FALSE; /* last event caused an error? */
  wmCurrMbox = NULL; /* the open mailbox */
  wmBoxMsgs = 0; /* the messages in it */
}


/* main...duh */
int main() {
  CONSOLE_SCREEN_BUFFER_INFO csbi;
  CONSOLE_CURSOR_INFO cci;
  DWORD conMode;
  INPUT_RECORD inBuf;
  DWORD eventsRead;
  DWORD err;


  /* Fun Windows startup stuff for handling the console */

  /* it's in the example, so it must be good */
  /* "if X=0,Y=0, we ran from a GUI, else not" */
  GetConsoleScreenBufferInfo(GetStdHandle(STD_OUTPUT_HANDLE), &csbi);
  if (csbi.dwCursorPosition.X | csbi.dwCursorPosition.Y) /* either non-zero? */
  { /* if we didn't, then give ourselves a fresh one */
    FreeConsole();
    AllocConsole();
  }
  SetConsoleTitle("winmutt - initialising");

  /* get console handles and largest console size */
  wmConOut = GetStdHandle(STD_OUTPUT_HANDLE);
  wmConIn = GetStdHandle(STD_INPUT_HANDLE);

  /* set the modes nicely */
  GetConsoleMode(wmConIn, &conMode);
  conMode |= (ENABLE_PROCESSED_INPUT | ENABLE_WINDOW_INPUT | ENABLE_MOUSE_INPUT);
  /* have to handle the line input myself, since it'll go auto-scrolling otherwise */
  conMode &= ~(ENABLE_LINE_INPUT | ENABLE_ECHO_INPUT);
  SetConsoleMode(wmConIn, conMode);

  GetConsoleMode(wmConOut, &conMode);
  /* have to handle everything myself, otherwise it scrolls even though it shouldn't */
  conMode &= ~(ENABLE_PROCESSED_OUTPUT | ENABLE_WRAP_AT_EOL_OUTPUT);
  SetConsoleMode(wmConOut, conMode);

  /* disable the cursor */
  GetConsoleCursorInfo(wmConOut, &cci);
  cci.bVisible = FALSE;
  SetConsoleCursorInfo(wmConOut, &cci);

  /* anyway, on with the program */
  /* get the screen size for startup messages being produced */
  GetConsoleScreenBufferInfo(wmConOut, &csbi);
  wmConSizeX = csbi.dwSize.X;
  wmConSizeY = csbi.dwSize.Y;

  initVars();

  wm_readConfig();

  if(!wm_parseMbox(DEFAULT_MBOX)) wm_Exit(1);

  /* just in case someone's resized it *already* */
  GetConsoleScreenBufferInfo(wmConOut, &csbi);
  wmConSizeX = csbi.dwSize.X;
  wmConSizeY = csbi.dwSize.Y;


  /* set top status bar, and display mailbox */

  wm_cls();
  wm_showStatusOrKeys("q:Quit  d:Del  u:Undel  s:Save  m:Mail  r:Reply  g:Group  ?:Help", WMKEYS);

  wm_displayMbox(-1);

  /* main loop, then */
  while(!wmQuitting) {
    if(!ReadConsoleInput(wmConIn, &inBuf, 1, &eventsRead)) {
      err = GetLastError();
      printf("Oh, I can't read console input any more.  byebye. (%ld)\n", err);
      exit(1);
    }

    switch(inBuf.EventType) {
      case KEY_EVENT:
        if(inBuf.Event.KeyEvent.bKeyDown == 1)
          wm_processKey(&inBuf.Event.KeyEvent);
        break;
      case MOUSE_EVENT:
        if(inBuf.Event.MouseEvent.dwEventFlags != MOUSE_MOVED &&
           inBuf.Event.MouseEvent.dwButtonState != MOUSE_WHEELED) /* ie, a (double)-click */
          wm_processClick(&inBuf.Event.MouseEvent);
        break;
      case WINDOW_BUFFER_SIZE_EVENT:
        /* coo, someone resized the window */
        /* redraw the mailbox, or something */
        break;
    }
  }

  wm_Exit(0);

  return 0;
}