/* C */
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>

/* Windows */
#include <wtypes.h>

/* Local */
#include "winmutt.h"
#include "wmbox.h"
#include "wmutil.h"
#include "wmvars.h"


/* get a header line */
char *wm_getHdr(char **line, int *type) {
  char word[512];
  char *hdr;
  int i;

  *type = 0;
  if(**line == '\0') return NULL; /* empty line */

  i = sscanf(*line, "%[-A-Za-z:] ", word);
  if(i == 0) { /* well, that didn't work */
    if(**line && isspace(**line)) { *type = 2; hdr = *line; } /* space-starting line */
                             else hdr = NULL;  /* end of headers */
  } else {
    hdr = *line;
    *line += strlen(word) + 1;
    *(*line - 1) = '\0';
    *type = 1; /* type 1 - 'Blah: text */
  }

  return hdr;
}


/* get the name and address from a From: header */

void wm_parseFromHdr(char *hdr, struct wmMail *mailData) {
  char tmpData[256], tmpCh; /* assumptions, assumptions... */
  int i;
  char *fromAddr, *angleBracket;

  while(*hdr && isspace(*hdr)) hdr++; /* skip possible leading ws */

  if(*hdr == '"') { /* As in "My Name" <me@here> */
    i = sscanf(hdr, "\"%[^\"]\"%c", tmpData, &tmpCh);
    if(i != 2) /* no closing " */
      return;
    else {
      mailData->mailFromName = (char *)malloc(strlen(tmpData) + 1);
      strcpy(mailData->mailFromName, tmpData);
      hdr += strlen(tmpData) + 2; /* plus the "" */
      while(*hdr && isspace(*hdr)) hdr++; /* skip any ws */
    }
  }

  if(*hdr == '\0') return; /* shouldn't happen - invalid if so really */

  if((angleBracket = strchr(hdr, '<')) == NULL) { /* address has no surrounding <> */
    fromAddr = hdr;
    while(!isspace(*hdr)) hdr++; /* skip to end of address */
    *hdr = '\0'; /* it might be anyway, but it doesn't hurt
                    Anyway, the point is to lose any trailing ws */
    mailData->mailFromAddr = (char *)malloc(strlen(fromAddr) + 1);
    strcpy(mailData->mailFromAddr, fromAddr);
  } else { /* address is <%s> */
    /* name first? */
    if(*hdr != '<') {
      sscanf(hdr, "%[^<]", tmpData); /* there is a <, we know this already */
      hdr += strlen(tmpData); /* skip forward to address */

      /* remove trailing ws */
      i = strlen(tmpData) - 1;
      while(isspace(tmpData[i])) { tmpData[i] = '\0'; i--; }

      mailData->mailFromName = (char *)malloc(strlen(tmpData) + 1);
      strcpy(mailData->mailFromName, tmpData);
    }

    i = sscanf(hdr, "<%[^>]%c", tmpData, &tmpCh); /* <[anything that's not '>']> - call me lazy */
    if((i == 2) && (tmpCh == '>')) { /* must be valid */
      mailData->mailFromAddr = (char *)malloc(strlen(tmpData) + 1);
      strcpy(mailData->mailFromAddr, tmpData);
    }
  } /* all done. anything else is invalid anyway. */
}

/* read in the mailbox */
int wm_parseMbox(char *fileName) {
  FILE *mbox;
  struct stat fileInfo;
  long fSize, boxLine = 0;
  BOOL inHeaders = FALSE;
  int currMsg = 0, currLines = 0, i, trailLines = 0;

  fpos_t where;
  struct wmMail *newMail = NULL;
  char thisLine[256], thisFrom[128], tz[10];
  char *currHdr = NULL, *thisHdr, *upTo;
  int hdrType;

  char day[4], mon[4];
  int dayNo, hr, mn, sc, yr, tmDiff, isDst;
  struct tm tmTemp;
  time_t thisTime;

  mbox = fopen(fileName, "r");
  if(mbox == NULL) {
    sprintf(wmTextBuf, "Couldn't open mailbox %s : %s (errno=%d)", fileName, strerror(errno), errno);
    wm_dispError(wmTextBuf);
    return 0;
  }

  SetConsoleTitle("winmutt - opening mailbox");

  /* get size for percentage marker */
  fstat(fileno(mbox), &fileInfo);
  fSize = fileInfo.st_size;
  /* is it DST right now? */
  time(&thisTime);
  tmTemp = *localtime(&thisTime);
  isDst = tmTemp.tm_isdst;

  while(!feof(mbox)) {
    if(fgets(thisLine, 256, mbox) == NULL) {
      if(feof(mbox)) break; /* exit the while */

      sprintf(wmTextBuf, "Unexpected error reading mailbox %s, exiting: %s (errno=%d)\n", fileName, strerror(errno), errno);
      wm_dispError(wmTextBuf);

      wm_Exit(1);
    }

    boxLine++;

    if(!strncmp(thisLine, "From ", 5)) {
      if(currMsg) newMail->mailSize = currLines; /* close off this mail */

      currMsg++;

      fgetpos(mbox, &where);
      sprintf(wmTextBuf, "Reading %s... %d (%d%%)",  fileName, currMsg, (where * 100) / fSize);
      wm_showMessage(wmTextBuf, wmColours[WMMESSAGECOL].colourNum);

      tmDiff = 0; /* default time zone */

      if(sscanf(thisLine, "From %127s %3s %3s %d %d:%d:%d %d ", thisFrom, day, mon, &dayNo, &hr, &mn, &sc, &yr) < 8) {
        if(sscanf(thisLine, "From %127s %3s %3s %d %d:%d:%d %9s %d ", thisFrom, day, mon, &dayNo, &hr, &mn, &sc, tz, &yr) < 9) { /* time zone? */
          sprintf(wmTextBuf, "Malformed From line in mailbox %s, line %ld", fileName, boxLine);
          wm_dispError(wmTextBuf);
          break;
        }
        /* timezone was supplied - which one? */
        tmDiff = 99; /* 'zone invalid' value */

        if(!(!stricmp(tz, "GMT") || !stricmp(tz, "UTC") || !stricmp(tz, "+0000"))) {
          if(sscanf(tz, "%3d00", &tmDiff) != 1) { /* ie, one sign char, 2 digits */
            if(!stricmp(tz, "CET")) tmDiff = 1;
            if(!stricmp(tz, "EST")) tmDiff = -5;
          }
        } else tmDiff = 0;

        if(tmDiff == 99) {
          sprintf(wmTextBuf, "Unrecognised time zone '%s' in mailbox %s, line %ld", tz, fileName, boxLine);
          wm_dispError(wmTextBuf);
          break;
        }
      }
      tmTemp.tm_year = yr - 1900;
      tmTemp.tm_mon = wm_getMonNo(mon);
      tmTemp.tm_mday = dayNo;
      tmTemp.tm_hour = hr;
      tmTemp.tm_min = mn;
      tmTemp.tm_sec = sc;
      tmTemp.tm_isdst = isDst;

      thisTime = mktime(&tmTemp);
      if(thisTime == -1) {
        sprintf(wmTextBuf, "Invalid time format in mailbox %s, line %ld", fileName, boxLine);
        wm_dispError(wmTextBuf);
        break;
      }
      thisTime -= (3600 * tmDiff); /* undo the time zone offset */

      newMail = (struct wmMail *)malloc(sizeof(struct wmMail));
      newMail->mailPtr = where - strlen(thisLine) - 1;
      newMail->recvDate = thisTime;
      newMail->sentDate = 0;
      newMail->mailFromName = NULL;
      newMail->mailFromAddr = NULL;
      newMail->mailSubj = NULL;
      newMail->mailSize = 0;

      if(wmFirstMail == NULL) wmFirstMail = newMail; /* list is empty */
                         else wmLastMail->nextMail = newMail; /* add it to the end */

      wmLastMail = newMail; /* terminate list */
      newMail->nextMail = NULL;

      currLines = 0; trailLines = 0;
      inHeaders = TRUE;
    } else if(!currMsg) {
      sprintf(wmTextBuf, "%s is not a mailbox", fileName);
      wm_dispError(wmTextBuf);
      break;
    } else { /* A line, and not 'From ' */
      i = strlen(thisLine) - 1;
      while(thisLine[i] == '\r' || thisLine[i] == '\n') { thisLine[i] = '\0'; i--; }

      if(inHeaders) {
        upTo = thisLine;

        if((thisHdr = wm_getHdr(&upTo, &hdrType)) == NULL) {
          inHeaders = FALSE;
          if(newMail->mailFromAddr == NULL) { /* No 'From:' ? - use the 'From ' */
            newMail->mailFromAddr = (char *)malloc(strlen(thisFrom) + 1);
            strcpy(newMail->mailFromAddr, thisFrom);
          }
          if(strlen(thisLine)) currLines++;
        } else { /* interesting headers? */
          if(hdrType == 1) {
            if(currHdr != NULL) free(currHdr);
            currHdr = (char *)malloc(strlen(thisHdr) + 1);
            strcpy(currHdr, thisHdr);
          }
          /* and now we should be checking currHdr, and using what
             getHdr retrieved instead of upTo. which it does anyway */
          if(!stricmp(thisHdr, "Subject:")) {
            if(newMail->mailSubj == NULL) { /* only take the first subject */
              newMail->mailSubj = (char *)malloc(strlen(upTo) + 1);
              strcpy(newMail->mailSubj, upTo);
            }
          } else if(!stricmp(thisHdr, "From:")) {
              if(newMail->mailFromAddr == NULL) /* only take the first From */
                wm_parseFromHdr(upTo, newMail);
          } else if(!stricmp(thisHdr, "Date:")) {
              /* well, this'll be fun */
              newMail->sentDate = 123;
          }

        }
      } else { /* 'normal' line */
        if(*thisLine == 0) trailLines++; /* blank line - ignore if last thing before a From */
        else {
          currLines += trailLines + 1;
          trailLines = 0;
        }
      }
    }

    /* drop the end of lines */
    while(strlen(thisLine) == 255)
      if(fgets(thisLine, 256, mbox) == NULL) {
        sprintf(wmTextBuf, "Unexpected error reading mailbox %s, exiting: %s (errno=%d)\n", fileName, strerror(errno), errno);
        wm_dispError(wmTextBuf);

        wm_Exit(1);
      }
  }

  newMail->mailSize = currLines; /* last mail */

  fclose(mbox);

  if(wmCurrMbox != NULL) free(wmCurrMbox);
  wmCurrMbox = (char *)malloc(strlen(fileName)+1);
  strcpy(wmCurrMbox, fileName);

  wmBoxMsgs = currMsg;
  wmBoxSize = fSize;
  return 1;
}


/* draw it */

void wm_displayMbox(int startMsg) {
  char statusBar[81], msgsAndSortTxt[64], boxTextSize[20];
  char mailLine[81], dateTxt[7];
  unsigned int mbHowMuch;
  COORD newPos;
  DWORD written;
  int i, scrnful;
  struct wmMail *thisMail;

  int arrowOpt = wm_findOpt("arrow_cursor")->optionValue.optionNumValue;

  sprintf(wmTextBuf, "winmutt - index of %s", wmCurrMbox);
  SetConsoleTitle(wmTextBuf);


  /* bottom status bar */

  if(wmBoxSize > 1048576)
    sprintf(boxTextSize, "%.1fM", (float)wmBoxSize / 1048576);
  else
    sprintf(boxTextSize, "%.1fK", (float)wmBoxSize / 1024);

  sprintf(msgsAndSortTxt, " [Msgs: %d %s]---(%s)", wmBoxMsgs, boxTextSize, wm_getStrOpt(wm_findOpt("sort")));
  strcpy(statusBar,"WinMutt: ---------------------------------------------------------------(end)---");

  if(wmBoxMsgs <= (wmConSizeY - 3)) {
    memcpy(statusBar + 73, "all", 3);
    scrnful = wmBoxMsgs;
    startMsg = 1;
  }
  else {
    scrnful = wmConSizeY - 3;
    if(startMsg == -1) startMsg = wmBoxMsgs - scrnful + 1;
    else {
      if(startMsg >= (wmBoxMsgs - scrnful + 1))
        startMsg = wmBoxMsgs - scrnful + 1;
      else {
        sprintf(wmTextBuf, "%02d%%", ((startMsg + scrnful - 1) * 100) / wmBoxMsgs);
        memcpy(statusBar + 73, wmTextBuf, 3);
      }
    }
  }
  wmScrnTopMsg = startMsg;

  /* 80 - 'WinMutt: ' - '[Msgs: xxx]---(%s)' - '---(end)---' = space left over for filename */
  mbHowMuch = 80 - 9 - strlen(msgsAndSortTxt) - 11;
  if(strlen(wmCurrMbox) < mbHowMuch) mbHowMuch = strlen(wmCurrMbox);
  
  memcpy(statusBar + 9, wmCurrMbox, mbHowMuch);
  memcpy(statusBar + 9 + mbHowMuch, msgsAndSortTxt, strlen(msgsAndSortTxt));

  wm_showStatusOrKeys(statusBar, WMSTATUS);


  /* mail lines */

  if(wmBoxMsgs) {
    SetConsoleTextAttribute(wmConOut, wmColours[WMNORMALCOL].colourNum);
    newPos.X = 0;

    thisMail = wmFirstMail;
    for(i = 1; i < startMsg; i++)
      thisMail = thisMail->nextMail;

    /*
    nnnn flg Jan 01 -From address-- (size) Subject
    */

    for(i = 0; i < scrnful; i++) {
      strftime(dateTxt, 7, "%b %d", localtime(&thisMail->recvDate));
      sprintf(mailLine, "%s%4d     %s %-*.*s (%4d) %-*.*s", (arrowOpt == WMOPTVAL_YES) ? "  " : "", startMsg, dateTxt, 15, 15, (thisMail->mailFromName == NULL) ? thisMail->mailFromAddr : thisMail->mailFromName, thisMail->mailSize, (arrowOpt == WMOPTVAL_YES) ? 80 - 41 : 80 - 39, (arrowOpt == WMOPTVAL_YES) ? 80 - 41 : 80 - 39, thisMail->mailSubj);

      newPos.Y = i + 1;
      SetConsoleCursorPosition(wmConOut, newPos);
      WriteConsole(wmConOut, mailLine, strlen(mailLine), &written, NULL);

      startMsg++;
      thisMail = thisMail->nextMail;
    }
    wmScrnBotMsg = startMsg - 1;

    wmCurrLine = -1; /* you've just redrawn the whole screen anyway */
    wm_setHighlight(newPos.Y);
  }
}
