/* Windows */
#include <wtypes.h>
#include <wincon.h>

/* Local */
#include "winmutt.h"
#include "wmbox.h"
#include "wmutil.h"
#include "wmvars.h"

void wm_processKey(KEY_EVENT_RECORD *kEvent) {
  int tmpMsg;

  switch(kEvent->wVirtualKeyCode) {
    case VK_UP: /* up a line */
      if(wmWasErr) { wmWasErr = FALSE; wm_clearStatus(); }

      if(wmCurrLine > 1)
        wm_setHighlight(wmCurrLine - 1);
      else {
        if(wmScrnTopMsg == 1) {
          wm_dispError("You are on the first message");
          wmWasErr = TRUE;
        } else { /* shift up a screen */
          tmpMsg = wmScrnTopMsg + wmCurrLine - 1;

          wmScrnTopMsg -= (wmConSizeY - 3);
          if(wmScrnTopMsg < 1) wmScrnTopMsg = 1;
          wm_displayMbox(wmScrnTopMsg);
          wm_setHighlight(tmpMsg - wmScrnTopMsg);
        }
      }

      break;

    case VK_DOWN: /* down a line */
      if(wmWasErr) { wmWasErr = FALSE; wm_clearStatus(); }

      if(wmCurrLine <= (wmScrnBotMsg - wmScrnTopMsg))
        wm_setHighlight(wmCurrLine + 1);
      else {
        if(wmScrnBotMsg == wmBoxMsgs) {
          wm_dispError("You are on the last message");
          wmWasErr = TRUE;
        } else { /* shift down a screen */
          tmpMsg = wmScrnTopMsg + wmCurrLine;

          wmScrnTopMsg += (wmConSizeY - 3);
          wm_displayMbox(wmScrnTopMsg);
          wm_setHighlight(tmpMsg - wmScrnTopMsg + 1);
        }
      }
      break;

    case VK_LEFT: /* up a page */
      if(wmWasErr) { wmWasErr = FALSE; wm_clearStatus(); }

      if(wmScrnTopMsg == 1) {
        if(wmCurrLine == 1) {
          wm_dispError("You are on the first message");
          wmWasErr = TRUE;
        } else wm_setHighlight(1);
      } else {
        tmpMsg = wmScrnTopMsg + wmCurrLine - 1;
        tmpMsg -= (wmConSizeY - 3); /* change message no. by a pageful */

        wmScrnTopMsg -= (wmConSizeY - 3);
        if(wmScrnTopMsg < 1) wmScrnTopMsg = 1;
        wm_displayMbox(wmScrnTopMsg);
        wm_setHighlight(tmpMsg - wmScrnTopMsg + 1);
      }
      break;

    case VK_RIGHT: /* down a page */
      if(wmWasErr) { wmWasErr = FALSE; wm_clearStatus(); }

      if(wmScrnBotMsg == wmBoxMsgs) {
        if(wmCurrLine == (wmScrnBotMsg - wmScrnTopMsg + 1)) {
          wm_dispError("You are on the last message");
          wmWasErr = TRUE;
        } else wm_setHighlight(wmScrnBotMsg - wmScrnTopMsg + 1);
      } else {
        tmpMsg = wmScrnTopMsg + wmCurrLine - 1;
        tmpMsg += (wmConSizeY - 3); /* change message no. by a pageful */

        wmScrnTopMsg += (wmConSizeY - 3);
        wm_displayMbox(wmScrnTopMsg);
        wm_setHighlight(tmpMsg - wmScrnTopMsg + 1);
      }

      break;

    case 'Q': /* quit */
      wm_showMessage("Quitting...", wmColours[WMMESSAGECOL].colourNum);

      wmQuitting = TRUE;
      break;
  }
}


void wm_processClick(MOUSE_EVENT_RECORD *mEvent) {
  /* clear error condition, erase message */
  if(wmWasErr) {
    wmWasErr = FALSE;
    wm_showMessage("", wmColours[WMNORMALCOL].colourNum);
  }
}
