/* C */
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <malloc.h>

/* Local */
#include "winmutt.h"
#include "wmuttrc.h"
#include "wmutil.h"
#include "wmvars.h"

/* pick out a word and convert it to a constant command code */

int wm_getRCCom(char **line) {
  char word[512];
  int i;

  /* ignore comments */
  if(*line[0] == '#') return WMRC_COMMENT;

  i = sscanf(*line, "%s", word);
  if(i == 0) /* no fields found. impossible. I *think*... */
    return WMRC_UNRECOG;

  *line += strlen(word) + 1;
  *(*line - 1) = '\0'; /* terminate it.  why not? */

  if(!stricmp(word, "source")) return WMRC_SOURCE;
  if(!stricmp(word, "alias")) return WMRC_ALIAS;
  if(!stricmp(word, "set")) return WMRC_SET;
  if(!stricmp(word, "unset")) return WMRC_UNSET;
  if(!stricmp(word, "color")) return WMRC_COLOUR;
  if(!stricmp(word, "colour")) return WMRC_COLOUR;
  if(!stricmp(word, "my_hdr")) return WMRC_MYHDR;

  /* anything else is unrecognised */
  return WMRC_UNRECOG;
}


/* start from *line, skip whitespace, then scanf for a single word
   retWord is the start of the word, line is moved past the end of
   the word, and retWord is terminated with \0 */

char *wm_getRCWord(char **line) {
  char *retWord;
  char wordBuf[512];
  int i;

  retWord = *line;

  /* skip leading white space */
  while(**line && isspace(**line)) { retWord++; *line++; }

  i = sscanf(retWord, "%s", wordBuf);
  if(i <= 0) /* we must've hit the end of the line */
    return NULL;

  *line += strlen(wordBuf); /* points at first ws character */
  if(**line != '\0') { /* if this isn't the end of the line anyway */
    **line = '\0'; /* terminates retWord */
    (*line)++; /* point at next word (or ws) */
  }

  return retWord;
}


/* fetch the rest of the rc line.  except kill WS too */

char *wm_getRCLine(char *line) {
  int i;

  /* skip leading white space */
  while(*line && isspace(*line)) line++;

  if(*line) { /* ie, if not at eol */
    /* dump trailing white space */
    i = strlen(line) - 1;
    while(isspace(line[i])) { line[i] = '\0'; i--; }
  }

  if(*line) return line;
       else return NULL;
}


/* start from *line, skip whitespace, then scanf for a thing made up
   of letters, bypass any trailing whitespace, check for =
   retWord is the start of the word, line is moved past the end of
   the word, and retWord is terminated with \0 */

char *wm_getRCOptName(char **line, int *equals) {
  char *retWord;
  char wordBuf[512];
  int i;
  char *eow;

  retWord = *line;
  *equals = 0;

  /* skip leading white space */
  while(**line && isspace(**line)) { retWord++; (*line)++; }

  i = sscanf(retWord, "%[A-Za-z_]", wordBuf);
  if(i <= 0) /* we must've hit the end of the line */
    return NULL;

  /* skip *line over the word we just got */
  *line += strlen(wordBuf);
  eow = *line;

  while(**line && isspace(**line)) (*line)++; /* skip possible trailing ws */

  /* if we're not at = or eol now, then it's invalid */
  if(**line && **line != '=') {
    *equals = -1; /* -1 is an error flag.  just-because */

    /* if there *is* an equals, find it */
    while(**line && **line != '=') (*line)++;
    if(**line == '=') /* if there was one, terminate the word there */
      **line = '\0';

    /* Drop any trailing ws */
    i = strlen(retWord) - 1;
    while(isspace(retWord[i])) { retWord[i] = '\0'; i--; }

    return retWord; /* word terminator wasn't eol, = or space */
  }

  if(**line == '=')
    *equals = 1;

  if(**line) (*line)++; /* move to the next character if not eol */

  *eow = '\0'; /* terminates retWord */

  return retWord;
}


/* read a word, hope it's a colour, return number if so */
/* return what the word was either way */

char *wm_getRCColNum(char **line, int *number) {
  char *retWord;

  retWord = wm_getRCWord(line);
  *number = -1;

  if(retWord) {
    if(!stricmp(retWord, "attachment")) *number = WMATTACHCOL;
    if(!stricmp(retWord, "body")) *number = WMBODYCOL;
    if(!stricmp(retWord, "bold")) *number = WMBOLDCOL;
    if(!stricmp(retWord, "error")) *number = WMERRORCOL;
    if(!stricmp(retWord, "header")) *number = WMHEADERCOL;
    if(!stricmp(retWord, "hdrdefault")) *number = WMHDRDEFCOL;
    if(!stricmp(retWord, "index")) *number = WMINDEXCOL;
    if(!stricmp(retWord, "indicator")) *number = WMINDICATORCOL;
    if(!stricmp(retWord, "markers")) *number = WMMARKERSCOL;
    if(!stricmp(retWord, "message")) *number = WMMESSAGECOL;
    if(!stricmp(retWord, "normal")) *number = WMNORMALCOL;
    if(!stricmp(retWord, "quoted")) *number = WMQUOTEDCOL;
    if(!stricmp(retWord, "search")) *number = WMSEARCHCOL;
    if(!stricmp(retWord, "signature")) *number = WMSIGCOL;
    if(!stricmp(retWord, "status")) *number = WMSTATUSCOL;
    if(!stricmp(retWord, "tree")) *number = WMTREECOL;
    if(!stricmp(retWord, "underline")) *number = WMUNDERLINECOL;
  }

  return retWord;
}


/* read a word, hope it's a colour, return the colour number if so */
/* which - 1 = fg, 2 = bg */

char *wm_getRCCol(char **line, WORD *number, int which) {
  char *retWord, *chkBright;
  int bright = 0;

  retWord = wm_getRCWord(line);
  *number = 99; /* invalid colour */

  if(retWord) {
    chkBright = retWord;
    if(!strnicmp(retWord, "bright", 6)) {
      chkBright += 6;
      bright = 1;
    }

    if(!stricmp(chkBright, "black")) *number = (which == 1 ? WMCOL_FGBLACK : WMCOL_BGBLACK);
    if(!stricmp(chkBright, "red")) *number = (which == 1 ? WMCOL_FGRED : WMCOL_BGRED);
    if(!stricmp(chkBright, "green")) *number = (which == 1 ? WMCOL_FGGREEN : WMCOL_BGGREEN);
    if(!stricmp(chkBright, "yellow")) *number = (which == 1 ? WMCOL_FGYELLOW : WMCOL_BGYELLOW);
    if(!stricmp(chkBright, "blue")) *number = (which == 1 ? WMCOL_FGBLUE : WMCOL_BGBLUE);
    if(!stricmp(chkBright, "magenta")) *number = (which == 1 ? WMCOL_FGMAGENTA : WMCOL_BGMAGENTA);
    if(!stricmp(chkBright, "cyan")) *number = (which == 1 ? WMCOL_FGCYAN : WMCOL_BGCYAN);
    if(!stricmp(chkBright, "white")) *number = (which == 1 ? WMCOL_FGWHITE : WMCOL_BGWHITE);
    if(!stricmp(chkBright, "default")) *number = (which == 1 ? WMCOL_FGDEFAULT : WMCOL_BGDEFAULT);

    if(bright && *number != 99) *number = *number | (which == 1 ? WMCOL_FGBRIGHT : WMCOL_BGBRIGHT);
  }
  
  return retWord;
}

/* output warning (while processing rc file) */

void wm_warnRC(char *errMsg, int line, char *file) {
  char outBuf[512];

  sprintf(outBuf, errMsg, line, file);
  wm_writeString(outBuf);
}


/* process the given RC file */

int wm_processConfFile(char *fileName) {
  FILE *confFile;
  char thisLine[512]; /* 512 max. line length.  too lazy to malloc */
  char *upTo, *nextWord, *colName;
  int lineNo, i, com, eq, thisCol;
  int fileErr = 0;
  WORD fgCol, bgCol;
  struct wmOption *thisOpt;

  /* new things */
  struct wmAlias *newAlias;

  confFile = fopen(fileName, "r");
  if(confFile == NULL) {
    if(errno == ENOENT && !strcmp(fileName, CONFIG_FILE)) /* file not found is ok if base file */
      return 0;

    sprintf(wmTextBuf, "Couldn't open rc file %s : %s (errno=%d)", fileName, strerror(errno), errno);
    wm_dispError(wmTextBuf);
    return 1; /* exit anyway */
  }

  lineNo = 0;

  while(!feof(confFile)) {
    /* get a line */
    if(fgets(thisLine, 512, confFile) == NULL) {
      if(feof(confFile)) break; /* exit the while */

      sprintf(wmTextBuf, "Unexpected error reading wmuttrc, exiting: %s (errno=%d)\n", strerror(errno), errno);
      wm_dispError(wmTextBuf);
      wm_Exit(1);
    }

    lineNo++;
    upTo = thisLine; /* where we are on the line */
    
    /* skip leading white space */
    while(*upTo && isspace(*upTo)) upTo++;

    /* dump trailing white space */
    i = strlen(thisLine) - 1;
    while(isspace(thisLine[i])) { thisLine[i] = '\0'; i--; }

    if(*upTo) { /* ie, if there's still something left after removing
                       the leading/trailing whitespace */

      /* get the command */
      com = wm_getRCCom(&upTo);

      switch(com) {
        case WMRC_COMMENT: break; /* ignore */


        case WMRC_UNRECOG: /* something unrecognised */
          sprintf(wmTextBuf, "Unrecognised command '%s'", thisLine);
          strcat(wmTextBuf, " (line %d of file %s)");
          wm_warnRC(wmTextBuf, lineNo, fileName);
          fileErr++;
          break;


        case WMRC_SOURCE:
          nextWord = wm_getRCWord(&upTo); /* get filename */
          if(nextWord != NULL) {
            if(!stricmp(nextWord, fileName)) { /* sourcing a file from inside itself? */
              wm_warnRC("Can't source recursively (line %d of file %s)", lineNo, fileName);
              fileErr++;
            } else fileErr += wm_processConfFile(nextWord); /* process that file */
          } else { /* didn't get a file to read */
            wm_warnRC("Invalid or missing filename for 'source' command (line %d of file %s)", lineNo, fileName);
            fileErr++;
          }
          break;


          case WMRC_ALIAS:
            nextWord = wm_getRCWord(&upTo); /* get alias name */
            if(nextWord != NULL) {
              if((newAlias = wm_findAlias(nextWord)) == NULL) { /* create a new one, else overwrite */
                newAlias = (struct wmAlias *)malloc(sizeof(struct wmAlias));
                newAlias->aliasName = (char *)malloc(strlen(nextWord) + 1);
                newAlias->aliasAddr = NULL; /* to indicate this is new, for errors later */
                strcpy(newAlias->aliasName, nextWord);
              }

              /* then get alias address */
              nextWord = wm_getRCLine(upTo);
              if(nextWord != NULL) {
                if(newAlias->aliasAddr != NULL) free(newAlias->aliasAddr); /* overwriting existing alias */

                newAlias->aliasAddr = (char *)malloc(strlen(nextWord) + 1);
                strcpy(newAlias->aliasAddr, nextWord);

                if(wmFirstAlias == NULL) wmFirstAlias = newAlias; /* list is empty */
                                    else wmLastAlias->nextAlias = newAlias; /* add it to the end */

                wmLastAlias = newAlias; /* terminate list */
                newAlias->nextAlias = NULL;
              } else { /* no address supplied */
                sprintf(wmTextBuf, "No address supplied for alias '%s'", newAlias->aliasName);
                strcat(wmTextBuf, " (line %d of file %s)");
                wm_warnRC(wmTextBuf, lineNo, fileName);
                fileErr++;

                if(newAlias->aliasAddr == NULL) { /* ie, weren't trying to overwrite existing alias */
                  free(newAlias->aliasName);
                  free(newAlias);
                } /* otherwise leave existing one alone */
              }
            } else {
              wm_warnRC("No name supplied for alias (line %d of file %s)", lineNo, fileName);  /* no name supplied */
              fileErr++;
            }
            break;

          case WMRC_SET:
          case WMRC_UNSET:
            nextWord = wm_getRCOptName(&upTo, &eq);
            if(nextWord != NULL && eq != -1) {
              if(eq && com == WMRC_UNSET) {
                wm_warnRC("You can't set a value using unset (unset and = found on line %d of file %s)", lineNo, fileName);
                fileErr++;
              } else {
                thisOpt = wm_findOpt(nextWord);

                if(thisOpt != NULL) {
                  if((!eq || com == WMRC_UNSET) && (thisOpt->optionType == WMOPT_STR || thisOpt->optionType == WMOPT_INT)) {
                    sprintf(wmTextBuf, "Option '%s' %s", thisOpt->optionName, (com == WMRC_UNSET ? "cannot be unset" : "requires a value"));
                    strcat(wmTextBuf, " (line %d of file %s)");
                    wm_warnRC(wmTextBuf, lineNo, fileName);
                    fileErr++;
                  } else { /* set the value */
                    if(!eq) /* just a 'set' or 'unset' */
                      if(com == WMRC_SET)
                        thisOpt->optionValue.optionNumValue = WMOPTVAL_YES;
                      else
                        thisOpt->optionValue.optionNumValue = WMOPTVAL_NO;
                    else { /* there was a '=' */
                      nextWord = wm_getRCLine(upTo);
                      if(nextWord == NULL) { /* ie, set foo= */
                        switch(thisOpt->optionType) {
                          case WMOPT_STR:
                            wm_setStrOpt(thisOpt, "");
                            break;
                          case WMOPT_INT:
                            thisOpt->optionValue.optionNumValue = 0;
                            break;
                          case WMOPT_BOOL:
                          case WMOPT_ASK:
                            sprintf(wmTextBuf, "Must specify a value for option '%s'", thisOpt->optionName);
                            strcat(wmTextBuf, " (line %d of file %s)");
                            wm_warnRC(wmTextBuf, lineNo, fileName);
                            fileErr++;
                            break;
                        }
                      } else { /* actual value supplied */
                        /* strip surrounding quotes */
                        if(nextWord[0] == '"' && nextWord[strlen(nextWord)-1] == '"' ||
                           nextWord[0] == '\'' && nextWord[strlen(nextWord)-1] == '\'') {
                          nextWord++;
                          nextWord[strlen(nextWord)-1] = '\0';
                        }
                        switch(thisOpt->optionType) {
                          case WMOPT_STR:
                            wm_setStrOpt(thisOpt, nextWord);
                            break;
                          case WMOPT_INT:
                            thisOpt->optionValue.optionNumValue = atoi(nextWord);
                            break;
                          case WMOPT_BOOL:
                          case WMOPT_ASK:
                            i = -1;

                            if(!stricmp(nextWord, "y") || !stricmp(nextWord, "yes"))
                              i = WMOPTVAL_YES;
                            if(!stricmp(nextWord, "n") || !stricmp(nextWord, "no"))
                              i = WMOPTVAL_NO;
                            /* ask values only allowed for ask options */
                            if(thisOpt->optionType == WMOPT_ASK) {
                              if(!stricmp(nextWord, "ask-y") || !stricmp(nextWord, "ask-yes"))
                                i = WMOPTVAL_ASKYES;
                              if(!stricmp(nextWord, "ask-n") || !stricmp(nextWord, "ask-no"))
                                i = WMOPTVAL_ASKNO;
                            }

                            /* no valid option value found */
                            if(i == -1) {
                              sprintf(wmTextBuf, "The value '%s' given for option '%s' is not allowed (yes/no%s only)", nextWord, thisOpt->optionName, (thisOpt->optionType == WMOPT_ASK ? "/ask-yes/ask-no" : ""));
                              strcat(wmTextBuf, " (line %d of file %s)");
                              wm_warnRC(wmTextBuf, lineNo, fileName);
                              fileErr++;
                            } else /* ok - set it */
                              thisOpt->optionValue.optionNumValue = i;
                            break;
                        }
                      }
                    }
                  }
                } else { /* option didn't exist */
                  sprintf(wmTextBuf, "Unknown option '%s' supplied to %sset", nextWord, (com == WMRC_SET ? "" : "un"));
                  strcat(wmTextBuf, " (line %d of file %s)");
                  wm_warnRC(wmTextBuf, lineNo, fileName);
                  fileErr++;
                }
              }
            } else { /* errors - missing or syntactically invalid option */
              if(nextWord == NULL)
                sprintf(wmTextBuf, "Missing option for %sset", (com == WMRC_SET ? "" : "un"));
              else if(eq == -1)
                sprintf(wmTextBuf, "Invalid option '%s' supplied to %sset", nextWord, (com == WMRC_SET ? "" : "un"));

              strcat(wmTextBuf, " (line %d of file %s)");
              wm_warnRC(wmTextBuf, lineNo, fileName);
              fileErr++;
            }
            break;

        case WMRC_COLOUR:
          colName = wm_getRCColNum(&upTo, &thisCol);
          if(colName == NULL || thisCol == -1) {
            if(colName == NULL) sprintf(wmTextBuf, "Missing item for colour");
                           else sprintf(wmTextBuf, "Unknown item '%s' for colour", colName);
            strcat(wmTextBuf, " (line %d of file %s)");
            wm_warnRC(wmTextBuf, lineNo, fileName);
            fileErr++;
          } else {
            nextWord = wm_getRCCol(&upTo, &fgCol, 1);
            if(nextWord == NULL || fgCol == 99) {
              if(nextWord == NULL) sprintf(wmTextBuf, "Missing foreground colour for item '%s'", colName);
                              else sprintf(wmTextBuf, "Unknown foreground colour '%s' for item '%s'", nextWord, colName);
              strcat(wmTextBuf, " (line %d of file %s)");
              wm_warnRC(wmTextBuf, lineNo, fileName);
              fileErr++;
            } else {
              nextWord = wm_getRCCol(&upTo, &bgCol, 2);
              if(nextWord == NULL || bgCol == 99) {
                if(nextWord == NULL) sprintf(wmTextBuf, "Missing background colour for item '%s'", colName);
                                else sprintf(wmTextBuf, "Unknown background colour '%s' for item '%s'", nextWord, colName);
                strcat(wmTextBuf, " (line %d of file %s)");
                wm_warnRC(wmTextBuf, lineNo, fileName);
                fileErr++;
              } else {
                wmColours[thisCol].colourNum = fgCol | bgCol;

                /* pattern to match? */
                nextWord = wm_getRCLine(upTo);
                if(wmColours[thisCol].colPattern != NULL) free(wmColours[thisCol].colPattern);

                if(nextWord != NULL) {
                  wmColours[thisCol].colPattern = (char *)malloc(strlen(nextWord)+1);
                  strcpy(wmColours[thisCol].colPattern, nextWord);
                } else wmColours[thisCol].colPattern = NULL;
              }
            }
          }
          break;

        case WMRC_MYHDR:
          /* haven't done this bit yet */
          break;

        default: /* bad value returned from getRCCom */
          sprintf(wmTextBuf, "Internal program error - invalid value %d returned from wm_getRCCom", com);
          wm_dispError(wmTextBuf);
          exit(1);
      }

      /* drop the end of over-long lines */
      while(strlen(thisLine) == 511)
        if(fgets(thisLine, 512, confFile) == NULL) {
          sprintf(wmTextBuf, "Unexpected error reading wmuttrc, exiting: %s (errno=%d)\n", strerror(errno), errno);
          wm_dispError(wmTextBuf);

          wm_Exit(1);
        }
    }
  }

  fclose(confFile);

  return fileErr;
}


/* read RC files, starting with wmuttrc */
void wm_readConfig() {
  CONSOLE_SCREEN_BUFFER_INFO csbi;
  COORD curPos;
  DWORD written;
  int i, fileErr;

  SetConsoleTitle("winmutt - reading files");

  fileErr = wm_processConfFile(CONFIG_FILE);

  if(fileErr) {
    sprintf(wmTextBuf, "Error%s found in config file.  Press a key to continue.", (fileErr > 1 ? "s" : ""));
    wm_dispError(wmTextBuf);

    wm_getPress();

    /* clear away messages */
    GetConsoleScreenBufferInfo(wmConOut,&csbi);

    for(i=1; i<=csbi.dwCursorPosition.Y; i++) {
      curPos.X = 0; curPos.Y = i;
      FillConsoleOutputCharacter(wmConOut, ' ', csbi.dwSize.X, curPos, &written);
    }
  }
}
