/* C */
#include <string.h>
#include <malloc.h>
#include <stdlib.h>

/* Local */
#include "wmvars.h"
#include "wmutil.h"


/* rc configuration options and defaults */

#define WMOPTS(x,y) { x, WMOPT_STR, (int)"\001" y }

struct wmOption wmOptions[] = {
  WMOPTS("alias_file", "wmuttrc"),
  WMOPTS("alternates", ""),
  { "arrow_cursor", WMOPT_BOOL, WMOPTVAL_NO },
  { "confirmappend", WMOPT_BOOL, WMOPTVAL_YES },
  { "confirmcreate", WMOPT_BOOL, WMOPTVAL_YES },
  { "delete", WMOPT_ASK, WMOPTVAL_ASKYES },
  { "fast_reply", WMOPT_BOOL, WMOPTVAL_NO },
  { "include", WMOPT_ASK, WMOPTVAL_ASKYES },
  { "markers", WMOPT_BOOL, WMOPTVAL_YES },
  { "move", WMOPT_ASK, WMOPTVAL_ASKNO },
  WMOPTS("pop_host", ""),
  WMOPTS("pop_user", ""),
  WMOPTS("record", ""),
  { "save_empty", WMOPT_BOOL, WMOPTVAL_YES },
  { "save_name", WMOPT_BOOL, WMOPTVAL_NO },
  WMOPTS("signature", "wmsig"),
  WMOPTS("sort", "date-sent"),
  { "use_from", WMOPT_BOOL, WMOPTVAL_YES },
  { "view_lines", WMOPT_INT, 24 },
  { NULL }
};


/* locate an alias given the name */

struct wmAlias *wm_findAlias(char *name) {
  struct wmAlias *thisAlias;

  if(wmFirstAlias == NULL) return NULL;

  thisAlias = wmFirstAlias;
  while(thisAlias != NULL) {
    if(!strcmp(thisAlias->aliasName, name)) break;
    thisAlias = thisAlias->nextAlias;
  }

  return thisAlias;
}


/* locate an option given the name */

struct wmOption *wm_findOpt(char *name) {
  int i = 0;

  while(wmOptions[i].optionName != NULL && stricmp(wmOptions[i].optionName, name))
    i++;

  if(wmOptions[i].optionName != NULL) return &wmOptions[i];

  return NULL;
}


/* get a string option value */

char *wm_getStrOpt(struct wmOption *opt) {
  /* sanity check */
  if(opt->optionType != WMOPT_STR) {
    wm_dispError("Tried to get a string value from a non-string option!\n");
    exit(1);
  }
  if(opt->optionValue.optionStrValue[0] == '\001')
    return (opt->optionValue.optionStrValue) + 1;
  else
    return opt->optionValue.optionStrValue;
}

/* set a string option value */

void wm_setStrOpt(struct wmOption *opt, char *value) {
  /* sanity check */
  if(opt->optionType != WMOPT_STR) {
    wm_dispError("Tried to set a string value on a non-string option!\n");
    exit(1);
  }

  /* if it's using the initialised value, you can't free it */
  if(opt->optionValue.optionStrValue[0] != '\001')
    free(opt->optionValue.optionStrValue);

  opt->optionValue.optionStrValue = (char*)malloc(strlen(value)+1);
  strcpy(opt->optionValue.optionStrValue, value);
}
